// Based on Brian Ernst's code from here:
// http://leetnightshade.com/archives/131
// http://leetnightshade.com/archives/167

// All of these functions are small, and we typically want them to execute quickly.
// Therefore, they are all defined in this header file, instead of a separate Vector2.cpp file.
// Also, they are all given the "inline" keyword.
// Using a cpp would make inlining impossible.

// What you should not do is use 'using' statements in header files -- it's impossible to
// unuse something in a source file that includes it, and you shouldn't force people to
// add extra stuff to the global scope.
// If you need to use things from other namespaces in your headers, fully qualify every name.
// It can be a pain sometimes, but it's really the right thing to do.

#ifndef VECTOR2_H_INCLUDED
#define VECTOR2_H_INCLUDED

#include <cmath>
#include <limits>
#include <algorithm>

namespace Emil
{

typedef float Scalar;
typedef int Bool;

class Vector2
{
public:
	Scalar x;
	Scalar y;

	Vector2();
	Vector2(Scalar X, Scalar Y);
	void set(Scalar X, Scalar Y);
	void setToVeryLarge();
	bool isVeryLarge();

	Scalar 	getLength();
	Scalar 	getLengthSquared() const;
	Scalar 	getAngle();

	void normalize();
	void clamp(Scalar maxLength);
	void scaleTo(Scalar newLength);
	//void scaleBy(Scalar value); same as operator*=
	void rotateBy(Scalar angle);
	void rotateBy90();
	void rotateBy180();
	void rotateBy270();
	void rotateTo(Scalar newAngle);
	void boxScale(Vector2 box);
	void scaleToScreenEdge(Vector2 origin, Vector2 screenMax);

	static Vector2 constructFromPolar(Scalar L, Scalar A);
	static Vector2 normalized(Vector2 vect);
	static Vector2 clamped(Vector2 vect, Scalar maxLength);
	static Vector2 scaledTo(Vector2 vect, Scalar newLength);
	//static Vector2 scaledBy(Vector2 vect, Scalar newLength);  same as operator*
	static Vector2 rotatedBy(Vector2 vect, Scalar angle);
	static Vector2 rotatedBy90(Vector2 vect);
	static Vector2 rotatedBy180(Vector2 vect);
	static Vector2 rotatedBy270(Vector2 vect);
	static Vector2 rotatedTo(Vector2 vect, Scalar newAngle);
	static Vector2 boxScaled(Vector2 vect, Vector2 box);

	static Scalar dot(const Vector2 vect1, const Vector2 vect2);
	static Scalar distanceBetween(const Vector2 vect1, const Vector2 vect2);
	static Scalar distanceBetweenSquared(const Vector2 vect1, const Vector2 vect2);
	//static Vector2 cartesian(Scalar x, Scalar y);
	//static Vector2 polar(Scalar radius, Scalar angle);
/*
	TODO: more versions, for convenience
	Vector2 normalized();
	Vector2 clamped(Scalar maxLength);
	Vector2 scaledTo(Scalar newLength);
	Vector2 rotatedBy(Scalar angle);
	Vector2 rotatedTo(Scalar newAngle);

	Scalar distanceTo(const Vector2 vect2);
	Scalar distanceToSquared(const Vector2 vect2);
*/

	Vector2& operator+=(const Vector2& pVector);
	Vector2& operator-=(const Vector2& pVector);
	Vector2& operator*=(Scalar num);
	Vector2& operator/=(Scalar num);

	Bool operator!=(const Vector2& vect) const;
	Bool operator==(const Vector2& vect) const;

	friend Vector2 operator-(Vector2& vect);
	friend Vector2 operator+(const Vector2& vect1, const Vector2& vect2);
	friend Vector2 operator-(const Vector2& vect1, const Vector2& vect2);
	friend Vector2 operator*(const Vector2& vect1, Scalar value);
	friend Vector2 operator/(const Vector2& vect1, Scalar value);
	friend Vector2 operator*(const Vector2& vect1, const Vector2& vect2);
	friend Vector2 operator/(const Vector2& vect1, const Vector2& vect2);
};


// integer version, doesn't have as many methods as the above
// because if you want stuff like rotation, you tend to want floating-point too
class Vector2i
{
public:
	int x;
	int y;

	Vector2i();
	Vector2i(int X, int Y);
	void set(int X, int Y);
	void setToVeryLarge();
	bool isVeryLarge();

	Vector2i& operator+=(const Vector2i& pVector);
	Vector2i& operator-=(const Vector2i& pVector);
	Vector2i& operator*=(int num);
	Vector2i& operator/=(int num);

	Bool operator!=(const Vector2i& vect) const;
	Bool operator==(const Vector2i& vect) const;

	friend Vector2i operator-(Vector2i& vect);
	friend Vector2i operator+(const Vector2i& vect1, const Vector2i& vect2);
	friend Vector2i operator-(const Vector2i& vect1, const Vector2i& vect2);
	friend Vector2i operator*(const Vector2i& vect1, int value);
	friend Vector2i operator/(const Vector2i& vect1, int value);
	friend Vector2i operator*(const Vector2i& vect1, const Vector2i& vect2);
	friend Vector2i operator/(const Vector2i& vect1, const Vector2i& vect2);

	operator Vector2() const { return Vector2(this->x, this->y); }
};


inline Vector2::Vector2()
{
}

inline Vector2i::Vector2i()
{
}

inline Vector2::Vector2(Scalar X, Scalar Y)
	: x( X )
	, y( Y )
{
}

inline Vector2i::Vector2i(int X, int Y)
	: x( X )
	, y( Y )
{
}

inline void Vector2::set(Scalar X, Scalar Y)
{
	x = X;
	y = Y;
}

inline void Vector2i::set(int X, int Y)
{
	x = X;
	y = Y;
}

// this VeryLarge stuff is a bit ugly, but it's very useful sometimes
// for example when you have a list of Vector2s and want to mark some of them as "unused"
// but don't want to waste the extra memory of a separate status variable for each Vector2
inline void Vector2::setToVeryLarge()
{
//	x = std::numeric_limits<float>::max();
//	y = std::numeric_limits<float>::max();
	x = 10000000;
	y = 10000000;
}

inline void Vector2i::setToVeryLarge()
{
//	x = std::numeric_limits<int>::max();
//	y = std::numeric_limits<int>::max();
	x = 2147483647;
	y = 2147483647;
}

inline bool Vector2::isVeryLarge()
{
//	return (x == std::numeric_limits<float>::max() && y == std::numeric_limits<float>::max());
	return (x == 10000000 && y == 10000000);
}

inline bool Vector2i::isVeryLarge()
{
//	return (x == std::numeric_limits<int>::max() && y == std::numeric_limits<int>::max());
	return (x == 2147483647 && y == 2147483647);
}

inline Vector2 Vector2::constructFromPolar(Scalar L, Scalar A)
{
	Scalar X = cos(A) * L;
	Scalar Y = sin(A) * L;
	Vector2 temp(X, Y);
	return temp;
}

inline Scalar Vector2::getLength()
{
	return std::sqrt(x * x + y * y);
}

inline Scalar Vector2::getLengthSquared() const
{
	return x * x + y * y;
}

inline Scalar Vector2::getAngle()
{
	return std::atan2(y, x);
}

inline void Vector2::normalize()
{
	if(x == 0 && y == 0)
		return;

	Scalar length = getLength();
	x /= length;
	y /= length;
}

inline Vector2 Vector2::normalized(Vector2 vect)
{
	if(vect.x == 0 && vect.y == 0)
		return vect;

	Scalar length = vect.getLength();
	vect.x /= length;
	vect.y /= length;
	return vect;
}

inline void Vector2::clamp(Scalar maxLength)
{
	Scalar length = getLengthSquared();
	if(length > maxLength * maxLength)
	{
		length = std::sqrt(length);
		x *= maxLength / length;
		y *= maxLength / length;
	}
}

inline Vector2 Vector2::clamped(Vector2 vect, Scalar maxLength)
{
	Scalar length = vect.getLengthSquared();
	if(length > maxLength * maxLength)
	{
		length = std::sqrt(length);
		vect.x *= maxLength / length;
		vect.y *= maxLength / length;
	}
	return vect;
}

inline void Vector2::scaleTo(Scalar newLength)
{
	if(x == 0 && y == 0)
		return;

	Scalar oldLength = getLength();
	x *= newLength / oldLength;
	y *= newLength / oldLength;
}

inline Vector2 Vector2::scaledTo(Vector2 vect, Scalar newLength)
{
	if(vect.x == 0 && vect.y == 0)
		return vect;

	Scalar oldLength = vect.getLength();
	vect.x *= newLength / oldLength;
	vect.y *= newLength / oldLength;
	return vect;
}

inline void Vector2::rotateBy(Scalar angle)
{
	Scalar cosResult = std::cos(angle);
	Scalar sinResult = std::sin(angle);

	//Essentially, apply a 2x2 rotation matrix to the vector
	Scalar newX = x * cosResult - y * sinResult;
	Scalar newY = x * sinResult + y * cosResult;

	x = newX;
	y = newY;
}

inline Vector2 Vector2::rotatedBy(Vector2 vect, Scalar angle)
{
	Scalar cosResult = std::cos(angle);
	Scalar sinResult = std::sin(angle);

	//Essentially, apply a 2x2 rotation matrix to the vector
	Scalar newX = vect.x * cosResult - vect.y * sinResult;
	Scalar newY = vect.x * sinResult + vect.y * cosResult;

	vect.x = newX;
	vect.y = newY;
	return vect;
}

inline void Vector2::rotateBy90()
{
	Scalar newX = -y;
	Scalar newY = x;

	x = newX;
	y = newY;
}

inline void Vector2::rotateBy180()
{
	x = -x;
	y = -y;
}

inline void Vector2::rotateBy270()
{
	Scalar newX = y;
	Scalar newY = -x;

	x = newX;
	y = newY;
}

inline Vector2 Vector2::rotatedBy90(Vector2 vect)
{
	Scalar newX = -vect.y;
	Scalar newY = vect.x;

	vect.x = newX;
	vect.y = newY;
	return vect;
}

inline Vector2 Vector2::rotatedBy180(Vector2 vect)
{
	vect.x = -vect.x;
	vect.y = -vect.y;
	return vect;
}

inline Vector2 Vector2::rotatedBy270(Vector2 vect)
{
	Scalar newX = vect.y;
	Scalar newY = -vect.x;

	vect.x = newX;
	vect.y = newY;
	return vect;
}

inline void Vector2::rotateTo(Scalar newAngle)
{
	Scalar length = getLength();
	x = length * std::cos(newAngle);
	y = length * std::sin(newAngle);
}

inline Vector2 Vector2::rotatedTo(Vector2 vect, Scalar newAngle)
{
	Scalar length = vect.getLength();
	vect.x = length * std::cos(newAngle);
	vect.y = length * std::sin(newAngle);
	return vect;
}

inline void Vector2::boxScale(Vector2 box)
{
	// usage note: box.x and y must both be positive
	// either this vector's x or y must be non-zero
	Scalar boxAspect = box.x / box.y;
	if (!y || (x / y) > boxAspect) // the x should reach the box edge
	{
		Scalar absx = std::fabs(x);
		if (absx != box.x)
		{
			Scalar scale = box.x / absx;
			x *= scale;
			y *= scale;
		}
	}
	else // the y should reach the box edge
	{
		Scalar absy = std::fabs(y);
		if (absy != box.y)
		{
			Scalar scale = box.y / absy;
			x *= scale;
			y *= scale;
		}
	}
}


inline Vector2 Vector2::boxScaled(Vector2 vect, Vector2 box)
{
	// usage note: box.x and y must both be positive
	// either vect.x or y must be non-zero
	Scalar boxAspect = box.x / box.y;
	if (!vect.y || (vect.x / vect.y) > boxAspect) // the x should reach the box edge
	{
		Scalar absx = std::fabs(vect.x);
		if (absx != box.x)
		{
			Scalar scale = box.x / absx;
			vect *= scale;
		}
	}
	else
	{
		Scalar absy = std::fabs(vect.y);
		if (absy != box.y)
		{
			Scalar scale = box.y / absy;
			vect *= scale;
		}
	}
	return vect;
}

inline void Vector2::scaleToScreenEdge(Vector2 origin, Vector2 screenMax)
{
	// screenMin is assumed to be (0,0)
	// this must be true: (0,0) < origin < screenMax
	// TODO: somehow make this function more general! feels kinda special-purpose right now
	Vector2 box;
	if (x > 0)
		box.x = screenMax.x - origin.x;
	else
		box.x = origin.x - 0;
	if (y > 0)
		box.y = screenMax.y - origin.y;
	else
		box.y = origin.y - 0;
	boxScale(box);
	(*this) += origin;
}

inline Scalar Vector2::dot(const Vector2 vect1, const Vector2 vect2)
{
	return vect1.x * vect2.x + vect1.y * vect2.y;
}

inline Scalar Vector2::distanceBetween(const Vector2 vect1, const Vector2 vect2)
{
	Vector2 temp = vect1 - vect2;
	return temp.getLength();
}

inline Scalar Vector2::distanceBetweenSquared(const Vector2 vect1, const Vector2 vect2)
{
	Vector2 temp = vect1 - vect2;
	return temp.getLengthSquared();
}


//	inline Vector2 Vector2::cartesian(Scalar radius,Scalar angle)
//	{
//		return Vector2(radius * std::cos(angle),radius * std::sin(angle));
//	}

//	inline Vector2 Vector2::polar(Scalar x,Scalar y)
//	{
//		return Vector2(std::atan2(y,x),std::sqrt(x * x + y * y));
//	}


inline Bool Vector2::operator!=(const Vector2& vector) const
{
	return x != vector.x || y != vector.y;
}

inline Bool Vector2i::operator!=(const Vector2i& vector) const
{
	return x != vector.x || y != vector.y;
}

inline Bool Vector2::operator==(const Vector2& vect) const
{
	return (*this != vect) == false;
}

inline Bool Vector2i::operator==(const Vector2i& vect) const
{
	return (*this != vect) == false;
}

inline Vector2& Vector2::operator+=(const Vector2& pVector)
{
	x += pVector.x;
	y += pVector.y;
	return *this;
}

inline Vector2i& Vector2i::operator+=(const Vector2i& pVector)
{
	x += pVector.x;
	y += pVector.y;
	return *this;
}

inline Vector2& Vector2::operator-=(const Vector2& pVector)
{
	x -= pVector.x;
	y -= pVector.y;
	return *this;
}

inline Vector2i& Vector2i::operator-=(const Vector2i& pVector)
{
	x -= pVector.x;
	y -= pVector.y;
	return *this;
}

inline Vector2& Vector2::operator*=(Scalar num)
{
	x *= num;
	y *= num;
	return *this;
}

inline Vector2i& Vector2i::operator*=(int num)
{
	x *= num;
	y *= num;
	return *this;
}

inline Vector2& Vector2::operator/=(Scalar num)
{
	x /= num;
	y /= num;
	return *this;
}

inline Vector2i& Vector2i::operator/=(int num)
{
	x /= num;
	y /= num;
	return *this;
}

inline Vector2 operator+(const Vector2& vect1, const Vector2& vect2)
{
	return Vector2(vect1.x + vect2.x,vect1.y + vect2.y);
}

inline Vector2i operator+(const Vector2i& vect1, const Vector2i& vect2)
{
	return Vector2i(vect1.x + vect2.x,vect1.y + vect2.y);
}

inline Vector2 operator-(const Vector2& vect1, const Vector2& vect2)
{
	return Vector2(vect1.x - vect2.x,vect1.y - vect2.y);
}

inline Vector2i operator-(const Vector2i& vect1, const Vector2i& vect2)
{
	return Vector2i(vect1.x - vect2.x,vect1.y - vect2.y);
}

inline Vector2 operator-(Vector2& vect)
{
	return vect * -1;
}

inline Vector2i operator-(Vector2i& vect)
{
	return vect * -1;
}

inline Vector2 operator*(const Vector2& vect1, Scalar value)
{
	return Vector2(vect1.x * value,vect1.y * value);
}

inline Vector2i operator*(const Vector2i& vect1, int value)
{
	return Vector2i(vect1.x * value,vect1.y * value);
}

inline Vector2 operator/(const Vector2& vect1, Scalar value)
{
	return Vector2(vect1.x / value,vect1.y / value);
}

inline Vector2i operator/(const Vector2i& vect1, int value)
{
	return Vector2i(vect1.x / value,vect1.y / value);
}

inline Vector2 operator*(const Vector2& vect1, const Vector2& vect2)
{
	return Vector2(vect1.x * vect2.x,vect1.y * vect2.y);
}

inline Vector2i operator*(const Vector2i& vect1, const Vector2i& vect2)
{
	return Vector2i(vect1.x * vect2.x,vect1.y * vect2.y);
}

inline Vector2 operator/(const Vector2& vect1, const Vector2& vect2)
{
	return Vector2(vect1.x / vect2.x,vect1.y / vect2.y);
}

inline Vector2i operator/(const Vector2i& vect1, const Vector2i& vect2)
{
	return Vector2i(vect1.x / vect2.x,vect1.y / vect2.y);
}


// When t = 0 returns from.
// When t = 1 returns to.
// When t = 0.5 returns the point midway between from and to.
inline Vector2 Lerp(Vector2 from, Vector2 to, float t)
{
	return (from * (1-t)) + (to * t);
}


inline Vector2 closestPointOnLine(Vector2 C, Vector2 A, Vector2 B)
{
	// Let the point be C (Cx,Cy) and the line be AB (Ax,Ay) to (Bx,By)
	// Let P be the point of perpendicular projection of C on AB

	float LSquared = Vector2::distanceBetweenSquared(A, B);

	if (LSquared < 0.00001f)
		return A;

	// r = ( AC dot AB ) / ( ||AB|| ^ 2 )
	// dot(U, V) = (Ux * Vx) + (Uy * Vy)
	float r = ((C.x-A.x)*(B.x-A.x) + (C.y-A.y)*(B.y-A.y)) / LSquared;
	// The parameter r indicates P's position along AB

	if (r <= 0)
	{
		// r=0      P = A
		// r<0      P is on the backward extension of AB
		return A;
	}
	if (r >= 1)
	{
		// r=1      P = B
		// r>1      P is on the forward extension of AB
		return B;
	}

	// 0<r<1    P is interior to AB
	// The point P can then be found:
	Vector2 P = A + ((B - A)*r);
	return P;
}


inline float closestDistanceBetweenPointAndLine(Vector2 C, Vector2 A, Vector2 B)
{
	// http://www.faqs.org/faqs/graphics/algorithms-faq/
	// Subject 1.02: How do I find the distance from a point to a line?

	// Let the point be C (Cx,Cy) and the line be AB (Ax,Ay) to (Bx,By)
	// Let P be the point of perpendicular projection of C on AB

	//L = std::sqrt( (Bx-Ax)^2 + (By-Ay)^2 )
	float LSquared = Vector2::distanceBetweenSquared(A, B);
	//float L = std::sqrt(LSquared);

	if (LSquared < 0.00001f)
		return Vector2::distanceBetween(A, C);

	float r = ((C.x-A.x)*(B.x-A.x) + (C.y-A.y)*(B.y-A.y)) / LSquared;
	// The parameter r indicates P's position along AB

	if (r <= 0)
	{
		// r=0      P = A
		// r<0      P is on the backward extension of AB
		return Vector2::distanceBetween(A, C);
	}
	if (r >= 1)
	{
		// r=1      P = B
		// r>1      P is on the forward extension of AB
		return Vector2::distanceBetween(B, C);
	}

	// 0<r<1    P is interior to AB
	// The point P can then be found:
	//Vector2 P;
	//P = A + ((B - A)*r);

	// Use another parameter s to indicate the location along PC, with the following meaning:
	// s<0      C is left of AB
	// s>0      C is right of AB
	// s=0      C is on AB

	//float s = ((A.y-C.y)*(B.x-A.x) - (A.x-C.x)*(B.y-A.y)) / LSquared;
	//return std::fabs(s) * std::sqrt(LSquared);
	// the above is the original code,
	// but i think it can be optimized to
	//return abs(aycyasdfdsfa / (L*L)) * L;
	//return (abs(aycyasdfdsfa) / (L*L)) * L;
	//return abs(aycyasdfdsfa) / L;
	float s = ((A.y-C.y)*(B.x-A.x) - (A.x-C.x)*(B.y-A.y));
	return std::fabs(s) / std::sqrt(LSquared);

}


inline float closestDistanceSquaredBetweenPointAndLine(Vector2 C, Vector2 A, Vector2 B)
{
	// faster than closestDistanceBetweenPointAndLine, useful when you only want to compare
	return Vector2::distanceBetweenSquared(C, closestPointOnLine(C, A, B));
}


void LineToLinearEquation(Vector2 p0, Vector2 p1, float& slope, float& offset, bool& isVertical)
{
	float divisor = p0.x - p1.x;
	if (divisor)
	{
		slope = (p0.y - p1.y) / divisor;
		offset = p0.y - (slope * p0.x);
		isVertical = false;
	}
	else
	{
		slope = 0;
		isVertical = true;
		offset = p0.x; // here it has a different meaning
	}
}


// this one assumes that the lines DO intersect, does not handle special cases
Vector2 FindIntersectionPoint(float line1offset, float line1slope, float line2offset, float line2slope)
{
	Vector2 result;
	result.x = -(line1offset - line2offset) / (line1slope - line2slope);
	result.y = line1slope * result.x + line1offset;
	return result;
}


// this one also handles special cases
int FindIntersectionPoint(Vector2& point, float line1offset, float line1slope, float line1vertical, float line2offset, float line2slope, float line2vertical)
{
	// return values:
	// 0 - the lines are parallel and never intersects
	// 1 - the lines intersect at one point
	// 2 - the lines are parallel and intersects forever

	if (line2vertical)
	{
		if (line1vertical) // the lines are parallel
		{
			if (line1offset == line2offset)
				return 2;
			else
				return 0;
		}
		else
		{
			float x = line2offset;
			float y = line1slope * x + line1offset;
			point.x = x;
			point.y = y;
		}
	}
	else
	{
		if (line1vertical)
		{
			float x = line1offset;
			float y = line2slope * x + line2offset;
			point.x = x;
			point.y = y;
		}
		else
		{
			if (line1slope == line2slope) // the lines are parallel
			{
				if (line1offset == line2offset)
					return 2;
				else
					return 0;
			}
			else
			{
				point = FindIntersectionPoint(line1offset, line1slope, line2offset, line2slope);
			}
		}
	}

	return 1;
}


int LineSegmentsIntersect(Vector2 l0p0,Vector2 l0p1, Vector2 l1p0,Vector2 l1p1)
{
	// return values:
	// 0 - the segments don't intersect
	// 1 - the segments intersect at one point
	// TODO: 2 - the segments are parallel and intersect at more than one point. NOT YET IMPLEMENTED!!

	// we have two line segments, defined by two endpoints each
	// we need to transform them into linear equations of the form:
	// y = slope*x + offset

	float line0slope;
	float line0offset;
	bool line0vertical = false;
	LineToLinearEquation(l0p0, l0p1, line0slope, line0offset, line0vertical);

	float line1slope;
	float line1offset;
	bool line1vertical = false;
	LineToLinearEquation(l1p0, l1p1, line1slope, line1offset, line1vertical);

	Vector2 intrsctnPnt;
	int result = FindIntersectionPoint(intrsctnPnt, line0offset, line0slope, line0vertical, line1offset, line1slope, line1vertical);
	if (result != 1)
		return 0;

	// the equations are infinite, but we only want intersections that happen within the actual line segment

	if (line0vertical)
	{
		float lineTopY = std::min(l0p0.y, l0p1.y);
		float lineBotY = std::max(l0p0.y, l0p1.y);
		if (!(lineTopY <= intrsctnPnt.y && intrsctnPnt.y <= lineBotY))
			return 0;
	}
	else
	{
		float lineLeftX  = std::min(l0p0.x, l0p1.x);
		float lineRightX = std::max(l0p0.x, l0p1.x);
		if (!(lineLeftX <= intrsctnPnt.x && intrsctnPnt.x <= lineRightX))
			return 0;
	}

	if (line1vertical)
	{
		float lineTopY = std::min(l1p0.y, l1p1.y);
		float lineBotY = std::max(l1p0.y, l1p1.y);
		if (!(lineTopY <= intrsctnPnt.y && intrsctnPnt.y <= lineBotY))
			return 0;
	}
	else
	{
		float lineLeftX  = std::min(l1p0.x, l1p1.x);
		float lineRightX = std::max(l1p0.x, l1p1.x);
		if (!(lineLeftX <= intrsctnPnt.x && intrsctnPnt.x <= lineRightX))
			return 0;
	}

	return 1;
}


}

#endif // VECTOR2_H_INCLUDED
