#ifndef XIAOLINE_H_INCLUDED
#define XIAOLINE_H_INCLUDED

typedef struct ALLEGRO_BITMAP ALLEGRO_BITMAP;

namespace Emil
{

// simple, fast line, 1 pixel wide
// http://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm
void Bresenline(float x1, float y1, float x2, float y2, void(*drawFunc)(int x, int y, const bool horizontal, float fraction, void* pExtraArgs), void* pExtraArgs);

// anti-aliased line
// http://en.wikipedia.org/wiki/Xiaolin_Wu%27s_line_algorithm
void Xiaoline(float x1, float y1, float x2, float y2, void(*drawFunc)(int x, int y, const bool horizontal, float fraction, void* pExtraArgs), void* pExtraArgs);


void GrayDrawFunc(int x, int y, const bool horizontal, float fraction, void* pObject);

class AlphaLine // can be used for callback by Bresenline and Xiaoline
{
	/*
	example usage:

	ALLEGRO_BITMAP* bb = al_get_backbuffer(display);
	al_set_blender(ALLEGRO_ADD, ALLEGRO_ALPHA, ALLEGRO_INVERSE_ALPHA);
	al_lock_bitmap(bb, ALLEGRO_PIXEL_FORMAT_ANY, 0);

	AlphaLine xargs;
	xargs.alpha = 0.5f;
	Xiaoline(x, y, x2, y2, AlphaLine::Draw, &xargs);

	al_unlock_bitmap(bb);
	al_set_blender(ALLEGRO_ADD, ALLEGRO_ONE, ALLEGRO_ZERO);

	however, it's sort of slow... writing a specialized callback function can often be faster
	*/

	public:

	float r;
	float g;
	float b;
	float alpha;

	static void Draw(int x, int y, const bool horizontal, float fraction, void* pObject)
	{
		AlphaLine* myself = (AlphaLine*)pObject;
		myself->ReallyDraw(x, y, horizontal, fraction);
	}
	//static void Draw(int x, int y, const bool horizontal, float fraction, void* pObject);
	void ReallyDraw(int x, int y, const bool horizontal, float fraction);
};


// the above Bresenline uses a callback function. Sometimes that's too slow, we want truly inline code.
// I can't figure out a nice way to do it with templates and functors :(
// so resort to a macro:
// http://gcc.gnu.org/onlinedocs/cpp/Variadic-Macros.html
// it's error-prone and tricky! to debug read preprocessor output, do something like:
// "c:\Program Files\CodeBlocks\MinGW\bin\mingw32-g++.exe" -I E:\gamedev\inwork\_emillib -E E:\gamedev\inwork\open_eye_visuals\vector_field_drawing2\main.cpp > E:\preprocessor_output.cpp

#define DEFINE_BRESENLINE_FUNC(funcName, drawingCode, ...) \
void funcName(float x1, float y1, float x2, float y2, __VA_ARGS__) \
{ \
	bool steep = fabs(y2 - y1) > fabs(x2 - x1); \
	if (steep) \
	{ \
		swap(x1, y1); \
		swap(x2, y2); \
	} \
 \
	if (x1 > x2) \
	{ \
		swap(x1, x2); \
		swap(y1, y2); \
	} \
 \
	int dx = (int)x2 - (int)x1; \
	int dy = std::abs((int)y2 - (int)y1); \
	int error = -dx / 2; \
 \
	int ystep; \
	if ((int)y1 < (int)y2) \
		ystep = 1; \
	else \
		ystep = -1; \
 \
 	if (steep) \
	{ \
		int x = (int)y1; \
		for (int y = (int)x1; y <= (int)x2; y++) \
		{ \
			drawingCode; \
			error += dy; \
			if (error > 0) \
			{ \
				x += ystep; \
				error -= dx; \
			} \
		} \
	} \
	else \
	{ \
		int y = (int)y1; \
		for (int x = (int)x1; x <= (int)x2; x++) \
		{ \
			drawingCode; \
			error += dy; \
			if (error > 0) \
			{ \
				y += ystep; \
				error -= dx; \
			} \
		} \
	} \
}


}

#endif // XIAOLINE_H_INCLUDED
